"""
    pakala: EVM symbolic execution tool and vulnerability scanner.
    Copyright (C) 2018 Korantin Auguste

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import collections
import logging
import functools
import time
import itertools

import claripy

from pakala import analyzer
from pakala import env
from pakala import utils
from pakala.state import State

logger = logging.getLogger(__name__)


DEBUG_MARK_PATH = []


def is_function(state, function):
    return state.solver.satisfiable([state.env.calldata.read(0, 4) == function])


def with_new_env(state):
    """Return a new identical state, rooted in a new, independent environment."""
    assert state.solver.satisfiable()
    old_env = state.env
    new_env = old_env.clean_copy()
    state = state.copy()
    state.env = new_env
    state.replace(functools.partial(env.replace, old_env, new_env))
    state.replace(state.solver.regenerate_hash_symbols())

    for read_k, read_v in state.storage_read.items():
        new_v = claripy.BVS("storage[%s]" % read_k, 256)
        state.replace(lambda ast: ast.replace(read_v, new_v))

    assert state.solver.satisfiable()
    return state


class RecursiveAnalyzer(analyzer.BaseAnalyzer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Each element in the queue is a couple:
        # 0) State, a state made of the successive applications of all the state
        #    of the path, except the last one.
        # 1) list, path of states we applied.
        self.path_queue = collections.deque()

        # The dict in the queue before
        self.state_done = set()

        self.reference_states = []

    @property
    def address(self):
        return self.reference_states[0][0].env.address

    @property
    def caller(self):
        return self.reference_states[0][0].env.caller

    def _search_path(self, composite_state, path):
        logger.debug("Search path: %s", path)
        logger.debug("Composite state: %s", composite_state)

        # If we already encountered the same composite state with some other
        # path...
        if hash(composite_state) in self.state_done:
            return
        self.state_done.add(hash(composite_state))

        logger.debug("Check for bugs in composite state...")
        solver_bug = self.check_state(composite_state, path=path)
        if solver_bug is not None:
            return solver_bug

        # If we kill the contract, we can't make any more call!
        if path[-1].selfdestruct_to is not None:
            return

        # We have to downsize the used solver to free memory in z3, otherwise
        # they will collectively end up eating all the memory.
        composite_state.solver.downsize()

        # For each reference state, find the right one (with an unused env)
        # and add it to the queue.
        for reference_states in self.reference_states:
            for reference_state in reference_states:
                if all(s is not reference_state for s in path):
                    self.path_queue.append(
                        (composite_state.copy(), path + [reference_state])
                    )
                    break

    def _append_state(self, composite_state, state):
        # TODO: Simplify/split that function. A bit too complex.
        logger.debug(
            "_append_state: appending state %s\nto composite state %s",
            state.debug_string(),
            composite_state.debug_string(),
        )

        assert composite_state.selfdestruct_to is None

        composite_state.solver = composite_state.solver.combine([state.solver])
        # TODO: Is that better to include even if not strictly needed?
        # composite_state.storage_read.update(state.storage_read)

        if not composite_state.solver.satisfiable():
            return []
        assert state.solver.satisfiable()

        for call in state.calls:
            composite_state.calls.append(call)

        if state.selfdestruct_to is not None:
            composite_state.selfdestruct_to = state.selfdestruct_to

        # Resolve read/write

        def apply_read(r_key, r_val, composite_state):
            """Apply a read operation with (key, value) to the state."""
            composite_states_next = []

            # Here we consider the cases where it's possible to read something
            # we previously wrote to.
            not_overwritten_c = []
            for w_key, w_val in composite_state.storage_written.items():
                read_written = [r_key == w_key, r_val == w_val]
                not_overwritten_c.append(r_key != w_key)

                cs = composite_state.copy()
                cs.solver.add(read_written)
                if cs.solver.satisfiable():
                    composite_states_next.append(cs)
                    logger.debug(
                        "Found key read %s, corresponding to key written %s",
                        r_key,
                        w_key,
                    )

            # Is it not something we previously wrote to?
            composite_state.solver.add(not_overwritten_c)
            composite_state.solver.add(
                state.storage_read[r_key] == self._read_storage(state, r_key)
            )
            if composite_state.solver.satisfiable():
                composite_states_next.append(composite_state)

            return composite_states_next

        composite_states = [composite_state]
        for r_key, r_val in state.storage_read.items():
            composite_states = list(
                itertools.chain.from_iterable(
                    apply_read(r_key, r_val, composite_state)
                    for composite_state in composite_states
                )
            )

        for composite_state in composite_states:
            # Delete any storage_written at the same key in the composite
            # state.
            for c_key in list(composite_state.storage_written.keys()):
                for key in state.storage_written.keys():
                    if not composite_state.solver.satisfiable(
                        extra_constraints=[key != c_key]
                    ):
                        del composite_state.storage_written[c_key]
                        break

            for key, val in state.storage_written.items():
                composite_state.storage_written[key] = val

        logger.debug("_append_state: found states:")
        for composite_state in composite_states:
            logger.debug(composite_state.debug_string())

        return composite_states

    def check_states(self, states, timeout, max_depth):
        states = [state for state in states if state.is_interesting()]
        if not states:
            return

        # Each state must have its own independent environment
        assert not self.reference_states

        # For each state, a list of equivalent state, but each in a different
        # env so that they can be stacked together.
        self.reference_states = []
        for state in states:
            self.reference_states.append(
                [with_new_env(state) for _ in range(max_depth)]
            )
            # Add it to the paths to explore
            self.path_queue.append((State(), [self.reference_states[-1][0]]))

        # Recursive exploration
        last_path_len = 0
        time_start = time.process_time()
        while self.path_queue:
            initial_composite_state, path = self.path_queue.popleft()

            if len(path) > last_path_len:
                logger.log(
                    utils.INFO_INTERACTIVE,
                    "Now scanning paths of length %i.",
                    len(path),
                )
                last_path_len = len(path)
            if len(path) > max_depth:
                logger.debug("Over the max allowed depth, stopping.")
                return

            if DEBUG_MARK_PATH and all(
                is_function(s, f) for s, f in zip(path, DEBUG_MARK_PATH)
            ):
                logger.warning("DEBUG_MARK_PATH len %i", len(path))
                logger.warning("path: %s", path)
                breakpoint()

            new_composite_states = self._append_state(initial_composite_state, path[-1])

            for composite_state in new_composite_states:
                solver = self._search_path(composite_state, path)
                if solver is not None:
                    return composite_state, path, solver

            if timeout and time.process_time() - time_start > timeout:
                logger.debug("Timeout at depth %i, stopping.", len(path))
                return



import logging
import numbers

import claripy

from pakala import utils

from web3 import Web3


# We can load up to this many keys from the contract storage. More than that
# and we won't load them all, and read them lazily instead (which is less precise).
MAX_STORAGE_KEYS = 32

# When we cannot list the keys, we can always try these ones:
STORAGE_KEYS_WHEN_CANNOT_LIST = list(range(10))


logger = logging.getLogger(__name__)


class BaseAnalyzer(object):
    """Base class for an Analyzer.

    Child classes need to define `caller` and `address`.
    """

    def __init__(self, max_wei_to_send, min_wei_to_receive, block="latest"):
        self.web3 = Web3()
        self.web3.eth.defaultBlock = block
        self.max_wei_to_send = max_wei_to_send
        self.min_wei_to_receive = min_wei_to_receive

        self.actual_storage = None

        # Whether or not actual_storage is guaranteed to contain all the storage,
        # or just a subset of it. Will be False for contracts with a lot of keys
        # so that we cannot load them all.
        # For testing we can replace actual_storage with a dict so it's never
        # actually filled. In that case we can assume it's exhaustive.
        self.actual_storage_exhaustive = True

    @property
    def hex_addr(self):
        return self.web3.toChecksumAddress(
            utils.number_to_address(utils.bvv_to_number(self.address))
        )

    def _read_storage_key(self, key):
        return self.web3.toInt(self.web3.eth.getStorageAt(self.hex_addr, key))

    def _fill_actual_storage(self):
        try:
            storage_keys = [
                self.web3.toInt(hexstr=k)
                for k in self.web3.parity.listStorageKeys(
                    self.hex_addr, MAX_STORAGE_KEYS, None, self.web3.eth.defaultBlock
                )
            ]
        except Exception as e:
            # If we cannot list storage keys, let's read the beginning of the
            # space, and below we will mark that it's not exhaustive anyway.
            logger.warning(
                "Cannot list storage keys (%s). We will lose a bit of accuracy. "
                "Try to use a node that supports the parity_listStorageKeys RPC. ",
                e.__class__.__name__,
            )
            storage_keys = STORAGE_KEYS_WHEN_CANNOT_LIST
            self.actual_storage_exhaustive = False
        else:
            assert len(storage_keys) <= MAX_STORAGE_KEYS
            self.actual_storage_exhaustive = len(storage_keys) < MAX_STORAGE_KEYS

        self.actual_storage = {k: self._read_storage_key(k) for k in storage_keys}

        logger.info(
            "Loaded %i storage slots from the contract (%s). %i non-zero.",
            len(storage_keys),
            "exhaustive" if self.actual_storage_exhaustive else "non-exhaustive",
            sum(1 for v in self.actual_storage.values() if v != 0),
        )
        logger.debug("actual_storage: %r", self.actual_storage)

    def _read_storage(self, state, key):
        logger.debug("Reading storage %r" % key)

        if self.actual_storage is None:
            self._fill_actual_storage()

        # If our storage is not exhaustive, let's try to concretize the key and read the
        # corresponding storage directly.
        if not self.actual_storage_exhaustive:
            try:
                concrete_keys = state.solver.eval(key, 2)
            except claripy.errors.UnsatError as e:
                # We will lose accuracy, and assume that our actual_storage is exhaustive...
                logger.debug(
                    "Encountered an exception when resolving key %r: %r", key, e
                )
            else:
                for concrete_key in concrete_keys:
                    if concrete_key not in self.actual_storage:
                        self.actual_storage[concrete_key] = self._read_storage_key(
                            concrete_key
                        )
                if len(concrete_keys) == 1:
                    return self.actual_storage[concrete_keys[0]]
                else:
                    # We will lose accuracy, and assume that our actual_storage is exhaustive...
                    logger.debug(
                        "Non-exhaustive storage and multiple values possible for key %r",
                        key,
                    )

        symbolic_storage = utils.bvv(0)  # When uninitialized: 0
        for k, v in self.actual_storage.items():
            if v != 0:
                symbolic_storage = claripy.If(key == k, v, symbolic_storage)

        return symbolic_storage

    def check_state(self, state, path=None):
        """Check a reachable state for bugs"""
        logger.debug("Check state: %s", state)
        logger.debug("Constraints: %s", state.solver.constraints)

        solver = state.solver.branch()

        if path is None:
            path = [state]
            # Static read were we never wrote, but we know the key is not symbolic.
            # So we go and fetch it.
            for key, value in state.storage_read.items():
                constraint = state.storage_read[key] == self._read_storage(state, key)
                solver.add(constraint)
                logger.debug("Add storage constraint: %s", constraint)

        for s in path:
            solver.add(list(s.env.extra_constraints()))
            solver.add(
                [
                    s.env.caller == utils.DEFAULT_CALLER,
                    s.env.origin == utils.DEFAULT_CALLER,
                ]
            )

        # Calls
        total_sent = sum(s.env.value for s in path)
        sent_constraints = [s.env.value < self.max_wei_to_send for s in path]

        total_received_by_me = utils.bvv(0)
        total_received = utils.bvv(0)

        for call in state.calls:
            # TODO: Improve delegatecall support! And make it clearer it's
            # delegatecall, not just based on the length.
            assert 6 <= len(call) <= 7
            value, to, gas = call[-3:]  # pylint: disable=unused-variable,invalid-name

            delegatecall = len(call) == 6

            if delegatecall:
                if solver.satisfiable(
                    extra_constraints=[to[159:0] == self.caller[159:0]]
                ):
                    logger.info("Found delegatecall bug.")
                    solver.add(to[159:0] == self.caller[159:0])
                    return solver
            else:
                total_received_by_me += claripy.If(
                    to[159:0] == self.caller[159:0], value, utils.bvv(0)
                )
                total_received += value
                solver.add(value <= total_sent + path[0].env.balance)

        final_balance = path[0].env.balance + total_sent - total_received

        # Suicide
        if state.selfdestruct_to is not None:
            constraints = [
                final_balance >= self.min_wei_to_receive,
                state.selfdestruct_to[159:0] == self.caller[159:0],
            ]
            logger.debug("Check for selfdestruct bug with constraints %s", constraints)
            if solver.satisfiable(extra_constraints=constraints):
                logger.info("Found selfdestruct bug.")
                solver.add(constraints)
                return solver

        if total_received_by_me is utils.bvv(0):
            return

        logger.debug("Found calls back to caller: %s", total_received_by_me)

        solver.add(sent_constraints)
        solver.add(
            [
                claripy.SGE(final_balance, 0),
                total_received_by_me > total_sent,  # I get more than what I sent?
                total_received_by_me > self.min_wei_to_receive,
            ]
        )

        if solver.satisfiable():
            logger.info("Found call bug.")
            return solver


class Analyzer(BaseAnalyzer):
    """Simple Analyzer class, where caller and address are given explicitly."""

    def __init__(self, address, caller, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.caller = caller
        self.address = address



import logging
import pprint
import json

from pakala import memory
from pakala import utils

logger = logging.getLogger(__name__)


class State(object):
    """Represents a state during the execution of a contract.
    It also contains the interactions with the world.
    """

    def __init__(self, env=None):
        self.env = env
        self.pc = 0  # pylint:disable=invalid-name
        self.stack = []

        # The "depth" we needed to reach that state. Basically how many branches
        # we executed to reach it (can be increased more if we got there by fuzzing).
        self.depth = 0

        self.memory = memory.Memory()

        # That's an override to the storage in the blockchain.
        # It's the storage that has been written at the end of the execution of
        # the contract.
        self.storage_written = {}

        # Storage read while executing the contract.
        self.storage_read = {}

        self.calls = []
        self.selfdestruct_to = None

        self.solver = utils.get_solver()

    def __repr__(self):
        return (
            "State(selfdestruct_to=%s, calls=%s, storage_written=%s, "
            "storage_read=%s, env=%s, solver=%s)"
        ) % (
            self.selfdestruct_to,
            self.calls,
            self.storage_written,
            self.storage_read,
            self.env,
            self.solver,
        )

    def _as_dict(self):
        return {
            "selfdestruct_to": self.selfdestruct_to,
            "calls": self.calls,
            "storage_written": self.storage_written,
            "storage_read": self.storage_read,
            "env": None if self.env is None else self.env.as_dict(),
            "solver": self.solver.as_dict(),
        }

    def debug_string(self):
        WIDTH = 90  # Arbitrary choice.
        try:
            return pprint.pformat(self._as_dict(), width=WIDTH)
        except Exception:
            # This can happen when pprint is trying to evaluate a claripy
            # object for pretty-printing (during sorted()) and there is no
            # simple way to avoid it.
            # Let's try to pprint what we can. Otherwise fallback to repr().
            s = "{\n"
            for k, v in self._as_dict().items():
                s += '"%s": ' % k
                try:
                    s += pprint.pformat(v, width=WIDTH)
                except Exception:
                    s += repr(v)
                s += ",\n"
            s += "}"
            return s

    def clean(self):
        """Clean the state, when it won't be executed anymore and we are only
        interested by the calls, selfdestructs..."""
        self.stack = []
        self.memory = memory.Memory()
        self.solver.downsize()

    def replace(self, r):
        """Call r() repeatedly, with every single AST that's present in the
        state:
            - in storage, read and written
            - calls
            - in selfdestruct data
            - in solver constraints
        r() can replace symbols in the AST by other symbols. Generally, r() is
        derived from Env.replace(), to substitute an environment with another.
        """
        logger.debug("State.replace %s", r)
        self.storage_written = {r(k): r(v) for k, v in self.storage_written.items()}
        self.storage_read = {r(k): r(v) for k, v in self.storage_read.items()}
        self.calls = [[r(i) for i in call] for call in self.calls]
        self.selfdestruct_to = (
            None if self.selfdestruct_to is None else r(self.selfdestruct_to)
        )
        self.solver.replace(r)

    def __hash__(self):
        l = [
            hash(self.env),
            hash(self.pc),
            hash(self.memory),
            hash(self.selfdestruct_to),
        ]
        for i in self.stack:
            l.append(hash(i))
        for call in self.calls:
            for arg in call:
                l.append(hash(arg))
        # The following is because the ordering shouldn't matter:
        x = 0
        for k, v in self.storage_written.items():
            x ^= hash((k, v))
        l.append(x)
        for k, v in self.storage_read.items():
            x ^= hash((k, v))
        l.append(x)
        for constraint in self.solver.constraints:
            x ^= hash(constraint)
        l.append(x)
        return hash(tuple(l))

    def stack_push(self, x):
        if len(self.stack) >= 1024:
            raise utils.CodeError("Stack overflow")
        self.stack.append(x)

    def stack_pop(self):
        if not self.stack:
            raise utils.CodeError("Stack underflow")
        return self.stack.pop()

    def is_interesting(self):
        return bool(
            self.storage_written or self.calls or self.selfdestruct_to is not None
        )

    def copy(self):
        """Make a shallow copy of the current environment. Needs to be fast."""
        new_state = State(self.env)
        new_state.pc = self.pc
        new_state.stack = self.stack[:]
        new_state.memory = self.memory.copy()
        new_state.storage_written = self.storage_written.copy()
        new_state.storage_read = self.storage_read.copy()
        new_state.solver = self.solver.branch()
        new_state.calls = self.calls[:]
        new_state.selfdestruct_to = self.selfdestruct_to
        new_state.depth = self.depth
        return new_state

    # TODO(palkeo): Get rid of that. Needed because of heapq in sm.py...
    def __eq__(self, other):
        return 1

    def __ne__(self, other):
        return 0

    __lt__ = __ne__
    __gt__ = __ne__
    __ge__ = __eq__
    __le__ = __eq__



import claripy
import unittest
from unittest.mock import patch
import logging

from pakala.analyzer import Analyzer
from pakala.env import Env
from pakala.state import State
from pakala import utils

from web3 import Web3

logging.basicConfig(level=logging.DEBUG)


class TestCheckState(unittest.TestCase):
    def setUp(self):
        self.env = Env(b"", caller=utils.DEFAULT_CALLER, address=utils.DEFAULT_ADDRESS)

        self.state = State(self.env)
        self.analyzer = Analyzer(
            address=self.env.address,
            caller=self.env.caller,
            max_wei_to_send=Web3.toWei(10, "ether"),
            min_wei_to_receive=Web3.toWei(1, "milliether"),
        )

    def check_state(self, state):
        return self.analyzer.check_state(state)

    def get_call(self, value, to=None):
        if to is None:
            to = self.env.caller
        return [
            utils.bvv(0),
            utils.bvv(0),
            utils.bvv(0),
            utils.bvv(0),
            value,
            to,
            utils.bvv(0),
        ]

    def get_delegatecall(self, to=None):
        if to is None:
            to = self.env.caller
        return [
            utils.bvv(0),
            utils.bvv(0),
            utils.bvv(0),
            utils.bvv(0),
            to,
            utils.bvv(0),
        ]

    def test_nothing(self):
        self.assertFalse(self.check_state(self.state))

    def test_selfdestruct(self):
        self.state.selfdestruct_to = self.env.caller
        self.assertTrue(self.check_state(self.state))

    def test_send_back(self):
        self.state.calls.append(self.get_call(self.env.value))
        self.assertFalse(self.check_state(self.state))

    def test_delegatecall(self):
        self.state.calls.append(self.get_delegatecall())
        self.assertTrue(self.check_state(self.state))

    def test_delegatecall_to_other(self):
        self.state.calls.append(self.get_delegatecall(to=utils.bvv(0)))
        self.assertFalse(self.check_state(self.state))

    def test_send_back_more(self):
        self.state.calls.append(self.get_call(self.env.value + Web3.toWei(1, "ether")))
        self.assertTrue(self.check_state(self.state))

    def test_send_back_if_impossible_block(self):
        self.state.calls.append(
            self.get_call(
                claripy.If(
                    self.env.block_number > 100000000000,
                    self.env.value + Web3.toWei(1, "ether"),
                    0,
                )
            )
        )
        self.assertFalse(self.check_state(self.state))

    def test_send_back_if_possible_block(self):
        self.state.calls.append(
            self.get_call(
                claripy.If(
                    self.env.block_number < 100000000000,
                    self.env.value + Web3.toWei(1, "ether"),
                    0,
                )
            )
        )
        self.assertTrue(self.check_state(self.state))

    def test_send_back_nothing(self):
        self.state.calls.append(self.get_call(utils.bvv(0)))
        self.assertFalse(self.check_state(self.state))

    def test_send_back_twice(self):
        self.state.calls.append(self.get_call(self.env.value))
        self.state.calls.append(self.get_call(self.env.value / 8))
        self.assertTrue(self.check_state(self.state))

    def test_send_back_fixed_amount(self):
        self.state.calls.append(self.get_call(Web3.toWei(1, "ether")))
        self.assertTrue(self.check_state(self.state))

    def test_send_back_to_someone_else(self):
        self.state.calls.append(
            self.get_call(Web3.toWei(1, "ether"), to=self.env.caller + 1)
        )
        self.assertFalse(self.check_state(self.state))

    def test_send_all(self):
        self.state.calls.append(self.get_call(self.env.balance))
        self.assertTrue(self.check_state(self.state))

    def test_send_back_calldata(self):
        self.state.calls.append(self.get_call(self.env.calldata.read(0, 32)))
        self.state.solver.add(claripy.UGT(self.env.calldata.read(0, 32), 0))
        self.assertTrue(self.check_state(self.state))

    def test_send_back_negative_signed(self):
        self.state.calls.append(self.get_call(self.env.calldata.read(0, 32)))
        self.state.solver.add(claripy.SLT(self.env.calldata.read(0, 32), 0))
        self.assertFalse(self.check_state(self.state))

    def test_send_back_negative_unsigned(self):
        self.state.calls.append(self.get_call(self.env.calldata.read(0, 32)))
        self.state.solver.add(claripy.ULT(self.env.calldata.read(0, 32), 0))
        self.assertFalse(self.check_state(self.state))

    # TODO: Fix it!
    @unittest.skip(
        "Known issue: we are sending back env.balance, "
        "that doesn't contain env.value, and it should!"
    )
    def test_send_all_and_selfdestruct(self):
        self.state.calls.append(self.get_call(self.env.balance, to=self.env.caller + 1))
        self.state.selfdestruct_to = self.env.caller
        self.assertFalse(self.check_state(self.state))

    def test_read_concrete(self):
        self.analyzer.actual_storage = {0: 0xBAD1DEA}

        self.state.storage_read[utils.bvv(0)] = claripy.BVS("storage[0]", 256)
        self.state.selfdestruct_to = self.state.storage_read[utils.bvv(0)]
        self.assertFalse(self.check_state(self.state))

        self.state.calls.append(
            self.get_call(
                Web3.toWei(1, "ether") * self.state.storage_read[utils.bvv(0)]
            )
        )
        self.assertTrue(self.check_state(self.state))

    def test_non_exhaustive_storage(self):
        self.analyzer.actual_storage = {1: 0xBAD1DEA}
        self.analyzer.actual_storage_exhaustive = False

        self.state.storage_read[utils.bvv(0)] = claripy.BVS("storage[0]", 256)
        self.state.selfdestruct_to = self.state.storage_read[utils.bvv(0)]

        # Suicide to storage[0] that contains our address (state.env.caller)
        with patch.object(self.analyzer, "_read_storage_key") as mock_read_storage_key:
            mock_read_storage_key.return_value = utils.bvv_to_number(
                self.state.env.caller
            )
            self.assertTrue(self.check_state(self.state))
            mock_read_storage_key.assert_called_with(0)

    def test_non_exhaustive_storage2(self):
        """Same as the previous test, but we suicide to 0 so it doesn't work."""
        self.analyzer.actual_storage = {1: 0xBAD1DEA}
        self.analyzer.actual_storage_exhaustive = False

        self.state.storage_read[utils.bvv(0)] = claripy.BVS("storage[0]", 256)
        self.state.selfdestruct_to = self.state.storage_read[utils.bvv(0)]

        # Same as above, but we suicide to 0 instead of caller.
        with patch.object(self.analyzer, "_read_storage_key") as mock_read_storage_key:
            mock_read_storage_key.return_value = 0
            self.assertFalse(self.check_state(self.state))
            mock_read_storage_key.assert_called_with(0)

    def test_exhaustive_storage(self):
        self.analyzer.actual_storage = {1: 0xBAD1DEA}
        self.analyzer.actual_storage_exhaustive = True

        self.state.storage_read[utils.bvv(0)] = claripy.BVS("storage[0]", 256)
        self.state.selfdestruct_to = self.state.storage_read[utils.bvv(0)]

        # Same as above, but we suicide to 0 instead of caller.
        with patch.object(self.analyzer, "_read_storage_key") as mock_read_storage_key:
            self.assertFalse(self.check_state(self.state))
            mock_read_storage_key.assert_not_called()



import unittest
from unittest import mock
import random
import numbers
import codecs

import claripy

from pakala.env import Env
from pakala import sm
from pakala import utils

# pylint: disable=undefined-variable
# flake8: noqa

from eth.vm.opcode_values import *


BVV_0 = utils.bvv(0)
BVV_1 = utils.bvv(1)


class TestSymbolicMachine(unittest.TestCase):
    """Basic tests for the members of the symbolic machine."""

    def setUp(self):
        code = codecs.decode(
            "6003600302600f56601b60006000a15b6101a5600060"
            "00a160019003801515600f57600660006000a1",
            "hex",
        )
        env = Env(code)

        self.sm = sm.SymbolicMachine(env)

    def test_add_branch(self):
        state = self.sm.branch_queue[0][1]
        state.pc += 1

        self.sm.add_branch(state)
        self.assertEqual(len(self.sm.branch_queue), 2)

        # Add duplicate branch
        self.sm.add_branch(state)
        self.assertEqual(len(self.sm.branch_queue), 2)
        self.assertEqual(len(self.sm.code_errors), 1)

        # Add unsatisfiable state
        state.solver.add(claripy.BVS("lol", 8) == 5)
        self.sm.add_branch(state)
        self.assertEqual(len(self.sm.branch_queue), 3)
        state.solver.add(claripy.BVV(1, 8) == claripy.BVV(0, 8))
        self.sm.add_branch(state)
        self.assertEqual(len(self.sm.branch_queue), 3)
        self.assertEqual(len(self.sm.code_errors), 1)

    def test_add_for_fuzzing(self):
        state = self.sm.branch_queue[0][1]
        variable = claripy.BVS("test", 8)
        state.solver.add(variable < 50)
        self.sm.add_for_fuzzing(
            state, variable, [min, max, None, None, None, None, 42, 1337]
        )
        self.assertTrue(
            any(s.solver.eval(variable, 2) == (0,) for _, s in self.sm.branch_queue)
        )
        self.assertTrue(
            any(s.solver.eval(variable, 2) == (49,) for _, s in self.sm.branch_queue)
        )
        self.assertTrue(
            any(s.solver.eval(variable, 2) == (42,) for _, s in self.sm.branch_queue)
        )
        self.assertGreater(len(self.sm.branch_queue), 3)

    def test_execute(self):
        self.sm.execute(timeout_sec=10)

    def test_get_coverage(self):
        # Just test that it seems to work
        self.assertEqual(self.sm.get_coverage(), 0)


class TestInstructions(unittest.TestCase):
    def assertBEqual(self, a, b):
        self.assertTrue((a == b).is_true(), msg="%s != %s" % (a, b))

    def run_code(self, code, env={}):
        code = bytes(code)
        self.sm = sm.SymbolicMachine(Env(code, **env))
        self.state = self.sm.branch_queue[0][1]
        return self.sm.exec_branch(self.state)

    def assert_stack(self, stack):
        self.assertEqual(len(stack), len(self.state.stack))
        for reference, observed in zip(stack, self.state.stack):
            self.assertBEqual(reference, observed)

    def test_jumpdest(self):
        self.assertTrue(self.run_code([JUMPDEST]))

    def test_add(self):
        self.run_code([PUSH1, 1, PUSH1, 2, ADD])
        self.assert_stack([3])
        self.run_code([PUSH1, 1, PUSH32] + [255] * 32 + [ADD])
        self.assert_stack([0])

    def test_sub(self):
        self.run_code([PUSH1, 2, PUSH1, 1, SUB])
        self.assert_stack([claripy.BVV(2 ** 256 - 1, 256)])

    def test_mul(self):
        self.run_code([PUSH1, 2, PUSH1, 4, MUL])
        self.assert_stack([8])

    def test_div(self):
        self.run_code([PUSH1, 2, PUSH1, 8, DIV])
        self.assert_stack([4])
        self.run_code([PUSH1, 3, PUSH1, 8, DIV])
        self.assert_stack([2])
        self.run_code([PUSH1, 10, PUSH1, 8, DIV])
        self.assert_stack([0])
        self.run_code([PUSH32] + [255] * 32 + [PUSH1, 1, DIV])
        self.assert_stack([0])
        self.run_code([PUSH1, 0, PUSH1, 1, DIV])
        self.assert_stack([0])

    def test_sdiv(self):
        self.run_code([PUSH32] + [255] * 32 + [PUSH1, 1, SDIV])
        self.assert_stack([2 ** 256 - 1])
        self.run_code([PUSH1, 0, PUSH1, 1, SMOD])
        self.assert_stack([0])

    def test_mod(self):
        self.run_code([PUSH1, 3, PUSH1, 7, MOD])
        self.assert_stack([1])
        self.run_code([PUSH1, 3, PUSH32] + [255] * 32 + [MOD])
        self.assert_stack([0])
        self.run_code([PUSH1, 0, PUSH1, 1, MOD])
        self.assert_stack([0])

    def test_smod(self):
        self.run_code([PUSH1, 3, PUSH32] + [255] * 32 + [SMOD])
        self.assert_stack([2 ** 256 - 1])
        self.run_code([PUSH1, 3, PUSH32] + [255] * 31 + [252, SMOD])
        self.assert_stack([2 ** 256 - 1])
        self.run_code([PUSH1, 3, PUSH1, 4, SMOD])
        self.assert_stack([1])
        self.run_code([PUSH1, 0, PUSH1, 1, SMOD])
        self.assert_stack([0])

    def test_addmod(self):
        self.run_code([PUSH1, 2, PUSH1, 1, PUSH1, 1, ADDMOD])
        self.assert_stack([0])
        self.run_code([PUSH1, 2, PUSH1, 2, PUSH1, 1, ADDMOD])
        self.assert_stack([1])
        self.run_code([PUSH1, 2, PUSH1, 2, PUSH1, 2, ADDMOD])
        self.assert_stack([0])
        self.run_code([PUSH1, 0, PUSH1, 2, PUSH1, 1, ADDMOD])
        self.assert_stack([0])

    def test_mulmod(self):
        self.run_code([PUSH1, 2, PUSH1, 1, PUSH1, 1, MULMOD])
        self.assert_stack([1])
        self.run_code([PUSH1, 2, PUSH1, 2, PUSH1, 1, MULMOD])
        self.assert_stack([0])
        self.run_code([PUSH1, 2, PUSH1, 2, PUSH1, 2, MULMOD])
        self.assert_stack([0])
        self.run_code([PUSH1, 0, PUSH1, 1, PUSH1, 1, MULMOD])
        self.assert_stack([0])

    def test_exp(self):
        self.run_code([PUSH1, 3, PUSH1, 7, EXP])
        self.assert_stack([343])

    def test_shl(self):
        self.run_code([PUSH1, 1, PUSH1, 1, SHL])
        self.assert_stack([2])
        self.run_code([PUSH1, 1, PUSH1, 0, SHL])
        self.assert_stack([1])
        self.run_code([PUSH1, 1, PUSH1, 2, SHL])
        self.assert_stack([4])
        self.run_code([PUSH1, 1, PUSH2, 1, 0, SHL])
        self.assert_stack([0])

    def test_shr(self):
        self.run_code([PUSH1, 2, PUSH1, 1, SHR])
        self.assert_stack([1])
        self.run_code([PUSH1, 2, PUSH1, 0, SHR])
        self.assert_stack([2])
        self.run_code([PUSH1, 2, PUSH1, 2, SHR])
        self.assert_stack([0])
        self.run_code([PUSH1, 2, PUSH2, 1, 0, SHR])
        self.assert_stack([0])

    def test_sar(self):
        self.run_code([PUSH1, 2, PUSH1, 1, SAR])
        self.assert_stack([1])
        self.run_code([PUSH1, 2, PUSH1, 0, SAR])
        self.assert_stack([2])
        self.run_code([PUSH1, 2, PUSH1, 2, SAR])
        self.assert_stack([2 ** 255])
        self.run_code([PUSH1, 2, PUSH2, 1, 0, SAR])
        self.assert_stack([2])

    def test_signextend(self):
        self.run_code([PUSH1, 0xFF, PUSH1, 0x0, SIGNEXTEND])
        self.assert_stack([2 ** 256 - 1])
        self.run_code([PUSH1, 0x3, PUSH1, 0x0, SIGNEXTEND])
        self.assert_stack([3])
        self.run_code([PUSH1, 0x0, PUSH1, 0x0, SIGNEXTEND])
        self.assert_stack([0])
        self.run_code([PUSH2, 0xFF, 0xFE, PUSH1, 0x1, SIGNEXTEND])
        self.assert_stack([2 ** 256 - 2])
        self.run_code([PUSH2, 0xFF, 0xFE, PUSH1, 0x2, SIGNEXTEND])
        self.assert_stack([0xFFFE])

    def test_lt(self):
        self.run_code([PUSH1, 7, PUSH1, 6, LT])
        self.assert_stack([BVV_1])
        self.run_code([PUSH1, 7, PUSH1, 7, LT])
        self.assert_stack([BVV_0])
        self.run_code([PUSH32] + [0xFF] * 31 + [250, PUSH32] + [0xFF] * 31 + [251, LT])
        self.assert_stack([BVV_0])
        self.run_code([PUSH32] + [0xFF] * 31 + [250, PUSH32] + [0x00] * 31 + [251, LT])
        self.assert_stack([BVV_1])

    def test_gt(self):
        self.run_code([PUSH1, 7, PUSH1, 6, GT])
        self.assert_stack([BVV_0])
        self.run_code([PUSH1, 7, PUSH1, 7, GT])
        self.assert_stack([BVV_0])
        self.run_code([PUSH32] + [0xFF] * 31 + [250, PUSH32] + [0xFF] * 31 + [251, GT])
        self.assert_stack([BVV_1])
        self.run_code([PUSH32] + [0xFF] * 31 + [250, PUSH32] + [0x00] * 31 + [251, GT])
        self.assert_stack([BVV_0])

    def test_slt(self):
        self.run_code([PUSH1, 7, PUSH1, 6, SLT])
        self.assert_stack([BVV_1])
        self.run_code([PUSH1, 7, PUSH1, 7, SLT])
        self.assert_stack([BVV_0])
        self.run_code([PUSH32] + [0xFF] * 31 + [250, PUSH32] + [0xFF] * 31 + [251, SLT])
        self.assert_stack([BVV_0])
        self.run_code([PUSH32] + [0xFF] * 31 + [250, PUSH32] + [0x00] * 31 + [251, SLT])
        self.assert_stack([BVV_0])

    def test_sgt(self):
        self.run_code([PUSH1, 7, PUSH1, 6, SGT])
        self.assert_stack([BVV_0])
        self.run_code([PUSH1, 7, PUSH1, 7, SGT])
        self.assert_stack([BVV_0])
        self.run_code([PUSH32] + [0xFF] * 31 + [250, PUSH32] + [0xFF] * 31 + [251, SGT])
        self.assert_stack([BVV_1])
        self.run_code([PUSH32] + [0xFF] * 31 + [250, PUSH32] + [0x00] * 31 + [251, SGT])
        self.assert_stack([BVV_1])

    def test_eq(self):
        self.run_code([PUSH1, 1, PUSH1, 1, EQ])
        self.assert_stack([BVV_1])
        self.run_code([PUSH1, 1, PUSH1, 2, EQ])
        self.assert_stack([BVV_0])

    def test_iszero(self):
        self.run_code([PUSH1, 1, ISZERO])
        self.assert_stack([BVV_0])
        self.run_code([PUSH1, 1, PUSH32] + 32 * [0xFF] + [ADD, ISZERO])
        self.assert_stack([BVV_1])

    def test_and(self):
        self.run_code([PUSH1, 3, PUSH1, 2, AND])
        self.assert_stack([2])

    def test_or(self):
        self.run_code([PUSH1, 3, PUSH1, 2, OR])
        self.assert_stack([3])
        self.run_code([PUSH1, 1, PUSH1, 2, OR])
        self.assert_stack([3])

    def test_xor(self):
        self.run_code([PUSH1, 3, PUSH1, 2, XOR])
        self.assert_stack([1])

        a = random.randint(0, 255)
        b = random.randint(0, 255)
        self.run_code([PUSH1, a, PUSH1, b, DUP1, XOR, XOR])
        self.assert_stack([a])

    def test_not(self):
        self.run_code([PUSH1, 0x43, NOT])
        self.assert_stack([(2 ** 256 - 1) ^ 0x43])
        self.run_code([PUSH1, 0, NOT])
        self.assert_stack([2 ** 256 - 1])
        self.run_code([PUSH32] + 32 * [0xFF] + [NOT])
        self.assert_stack([0])

        a = random.randint(0, 255)
        self.run_code([PUSH1, a, NOT, NOT])
        self.assert_stack([a])

    def test_byte(self):
        self.run_code([PUSH2, 0x42, 0x43, PUSH1, 0, BYTE])
        self.assert_stack([0])
        self.run_code([PUSH2, 0x42, 0x43, PUSH1, 29, BYTE])
        self.assert_stack([0])
        self.run_code([PUSH2, 0x42, 0x43, PUSH1, 30, BYTE])
        self.assert_stack([0x42])
        self.run_code([PUSH2, 0x42, 0x43, PUSH1, 31, BYTE])
        self.assert_stack([0x43])
        self.run_code([PUSH2, 0x42, 0x43, PUSH1, 32, BYTE])
        self.assert_stack([0])
        self.run_code([PUSH2, 0x42, 0x43, PUSH1, 50, BYTE])
        self.assert_stack([0])
        self.run_code([PUSH3, 0x41, 0x42, 0x43, PUSH1, 29, BYTE])
        self.assert_stack([0x41])

    def test_pc(self):
        self.run_code([PUSH1, 1, PC])
        self.assert_stack([1, 2])

    def test_gas(self):
        self.run_code([GAS])
        self.assert_stack([self.state.env.gas])

    def test_address(self):
        self.run_code([ADDRESS])
        self.assert_stack([self.state.env.address])

    def test_balance(self):
        self.run_code([ADDRESS, BALANCE], env={"address": claripy.BVV(789, 256)})
        self.assert_stack([self.state.env.balance])

        with self.assertRaises(ValueError):
            self.run_code([PUSH1, 1, BALANCE])

    def test_origin(self):
        self.run_code([ORIGIN])
        self.assert_stack([self.state.env.origin])

    def test_caller(self):
        self.run_code([CALLER])
        self.assert_stack([self.state.env.caller])

    def test_callvalue(self):
        self.run_code([CALLVALUE])
        self.assert_stack([self.state.env.value])

    def test_blockhash(self):
        self.run_code([NUMBER, BLOCKHASH])
        self.assert_stack([self.state.env.block_hashes[self.state.env.block_number]])

    def test_timestamp(self):
        self.run_code([TIMESTAMP])
        self.assert_stack([self.state.env.block_timestamp])

    def test_number(self):
        self.run_code([NUMBER])
        self.assert_stack([self.state.env.block_number])

    def test_coinbase(self):
        self.run_code([COINBASE])
        self.assert_stack([self.state.env.coinbase])

    def test_difficulty(self):
        self.run_code([DIFFICULTY])
        self.assert_stack([self.state.env.difficulty])

    def test_pop(self):
        self.run_code([PUSH1, 1, POP])
        self.assert_stack([])
        with self.assertRaises(utils.CodeError):
            self.run_code([POP])

    def test_jump(self):
        r = self.run_code([PUSH1, 4, JUMP, PC, JUMPDEST])
        self.assertFalse(r)
        self.assert_stack([])
        self.assertEqual(len(self.sm.branch_queue), 2)
        self.sm.exec_branch(self.sm.branch_queue[1][1])
        self.assert_stack([])

        with self.assertRaises(utils.CodeError):
            self.run_code([PUSH1, 4, JUMP, POP, PC])

    def test_jumpi(self):
        r = self.run_code([PUSH1, 42, CALLVALUE, EQ, PUSH1, 8, JUMPI, PC, JUMPDEST])
        self.assertFalse(r)

        def true_state(state):
            self.state = state
            self.sm.exec_branch(self.state)
            try:
                self.assert_stack([7])
                self.assertEqual(self.state.solver.min(self.state.env.value), 0)
            except AssertionError:
                return False
            return True

        def false_state(state):
            self.state = state
            self.sm.exec_branch(self.state)
            try:
                self.assert_stack([])
                self.assertEqual(self.state.solver.min(self.state.env.value), 1337)
            except AssertionError:
                return False
            return True

        # We don't know which state is first (it changes). So we do that
        # to avoid flakiness.
        self.assertTrue(any(true_state(bq[1])) for bq in self.sm.branch_queue)
        self.assertTrue(any(false_state(bq[1])) for bq in self.sm.branch_queue)

        with self.assertRaises(utils.CodeError):
            self.run_code([PUSH1, 0, PUSH1, 0, JUMPI])

    def test_push(self):
        self.run_code([PUSH1, 2])
        self.assert_stack([2])
        self.run_code([PUSH3, 2, 0, 0, PUSH2, 0, 2])
        self.assert_stack([2 * 65536, 2])

    def test_dup(self):
        self.run_code([PUSH1, 1, PUSH1, 2, PUSH1, 3, DUP3])
        self.assert_stack([1, 2, 3, 1])
        self.run_code([PUSH1, 1, PUSH1, 2, PUSH1, 3, DUP1])
        self.assert_stack([1, 2, 3, 3])

    def test_swap(self):
        self.run_code([PUSH1, 1, PUSH1, 2, PUSH1, 3, SWAP2])
        self.assert_stack([3, 2, 1])
        self.run_code([PUSH1, 1, PUSH1, 2, PUSH1, 3, SWAP1])
        self.assert_stack([1, 3, 2])

    def test_log(self):
        self.run_code([PUSH1, 0, PUSH1, 0, PUSH1, 0, LOG1])

    def test_sha3_empty(self):
        self.run_code([PUSH1, 0, PUSH1, 0, SHA3])
        self.assertEqual(1, len(self.state.stack))
        sha3, = self.state.solver.eval(self.state.stack[0], 1)
        self.assertEqual(
            hex(sha3),
            hex(0xC5D2460186F7233C927E7DB2DCC703C0E500B653CA82273B7BFAD8045D85A470),
        )

    def test_sha3_zeros(self):
        self.run_code([PUSH1, 32, PUSH1, 0, SHA3])
        self.assertEqual(1, len(self.state.stack))
        sha3, = self.state.solver.eval(self.state.stack[0], 1)
        self.assertEqual(
            hex(sha3),
            hex(0x290DECD9548B62A8D60345A988386FC84BA6BC95484008F6362F93160EF3E563),
        )

    def test_sha3_a_mstore8(self):
        self.run_code([PUSH1, 0x61, PUSH1, 0, MSTORE8, PUSH1, 1, PUSH1, 0, SHA3])
        self.assertEqual(1, len(self.state.stack))
        sha3, = self.state.solver.eval(self.state.stack[0], 1)
        self.assertEqual(
            hex(sha3),
            # sha3('a')
            hex(0x3AC225168DF54212A25C1C01FD35BEBFEA408FDAC2E31DDD6F80A4BBF9A5F1CB),
        )

    def test_sha3_a_mstore(self):
        self.run_code([PUSH1, 0x61, PUSH1, 0, MSTORE, PUSH1, 1, PUSH1, 31, SHA3])
        self.assertEqual(1, len(self.state.stack))
        sha3, = self.state.solver.eval(self.state.stack[0], 1)
        self.assertEqual(
            hex(sha3),
            # sha3('a')
            hex(0x3AC225168DF54212A25C1C01FD35BEBFEA408FDAC2E31DDD6F80A4BBF9A5F1CB),
        )

    def test_stop(self):
        r = self.run_code([STOP])
        self.assertTrue(r)

    def test_return(self):
        r = self.run_code([RETURN])
        self.assertTrue(r)

    def test_calldataload(self):
        self.run_code([PUSH1, 0, CALLDATALOAD, PUSH1, 0, CALLDATALOAD])
        self.assertTrue(self.state.stack[0] is self.state.stack[1])
        self.run_code([PUSH1, 0, CALLDATALOAD, PUSH1, 32, CALLDATALOAD])
        self.assertTrue(self.state.stack[0] is not self.state.stack[1])

    def test_calldatasize(self):
        self.run_code([CALLDATASIZE, PUSH1, 4, CALLDATALOAD])
        self.assertEqual(self.state.solver.min(self.state.stack[0]), 36)

    def test_calldatacopy(self):
        # size, dstart, mstart
        self.run_code(
            [
                PUSH1,
                64,
                PUSH1,
                20,
                PUSH1,
                10,
                CALLDATACOPY,
                PUSH1,
                10,
                MLOAD,
                PUSH1,
                42,
                MLOAD,
            ]
        )
        self.assert_stack(
            [self.state.env.calldata.read(20, 32), self.state.env.calldata.read(52, 32)]
        )

    def test_codesize(self):
        self.run_code([PUSH1, 0, POP, CODESIZE])
        self.assert_stack([4])

    def test_extcodesize_self(self):
        self.run_code([ADDRESS, EXTCODESIZE])
        self.assert_stack([2])

    def test_extcodesize_other(self):
        self.run_code([PUSH1, 42, EXTCODESIZE])
        self.assertTrue(len(self.state.stack), 1)
        self.assertTrue(
            self.state.solver.satisfiable(extra_constraints=[self.state.stack[0] == 0])
        )
        self.assertTrue(
            self.state.solver.satisfiable(
                extra_constraints=[self.state.stack[0] == 1337]
            )
        )

    def test_codecopy(self):
        code = [CODESIZE, PUSH1, 0, DUP1, CODECOPY, PUSH1, 0, MLOAD]
        self.run_code(code)
        self.assert_stack([bytes(code).ljust(32, b"\0")])

    def test_mload(self):
        self.run_code([PUSH1, 0, MLOAD, PUSH1, 0, MLOAD])
        self.assertTrue(self.state.stack[0] is self.state.stack[1])

    def test_mstore(self):
        self.run_code([PUSH1, 42, PUSH1, 1, MSTORE, PUSH1, 1, MLOAD])
        self.assert_stack([42])
        self.run_code([PUSH1, 42, PUSH1, 0, MSTORE, PUSH1, 0, MLOAD])
        self.assert_stack([42])
        self.run_code(
            [
                PUSH2,
                0xCA,
                0xFE,
                PUSH1,
                64,
                MSTORE,
                PUSH2,
                0xDE,
                0xAD,
                PUSH1,
                62,
                MSTORE,
                PUSH1,
                66,
                MLOAD,
            ]
        )
        self.assert_stack([0xDEADCAFE0000])

    def test_mstore8(self):
        self.run_code([PUSH2, 0x13, 0x37, PUSH1, 30, MSTORE8, PUSH1, 0, MLOAD])
        self.assert_stack([0x3700])

    def test_msize(self):
        self.run_code([PUSH1, 1, PUSH1, 0, MSTORE8, MSIZE])
        self.assert_stack([1])
        self.run_code([PUSH1, 1, PUSH1, 32, MSTORE, MSIZE])
        self.assert_stack([64])

    def test_invalid_opcode(self):
        with self.assertRaises(utils.CodeError):
            self.run_code([0x01, 0x02, 0x03, 0xFE, 0xFF])


class TestOutcomes(unittest.TestCase):
    def assertBEqual(self, a, b):
        self.assertTrue((a == b).is_true(), msg="%s != %s" % (a, b))

    def outcomes(self, code, env={}):
        code = bytes(code)
        self.sm = sm.SymbolicMachine(Env(code, **env))
        self.sm.execute(timeout_sec=10)
        return self.sm.outcomes

    def test_sload_symbolic(self):
        outcome1, outcome2 = self.outcomes(
            [PUSH1, 42, CALLVALUE, SSTORE, PUSH1, 0, SLOAD]
        )
        self.assertEqual(len(outcome1.storage_read), 0)
        self.assertEqual(len(outcome1.storage_written), 1)

        self.assertTrue(utils.bvv(0) in outcome2.storage_read)
        self.assertEqual(len(outcome2.storage_read), 1)

    def test_sstore_symbolic(self):
        outcome1, outcome2 = self.outcomes(
            [
                PUSH1,
                42,
                PUSH1,
                0,
                SSTORE,
                PUSH1,
                0,
                SLOAD,
                PUSH1,
                43,
                CALLVALUE,
                SSTORE,
                PUSH1,
                0,
                SLOAD,
            ]
        )

        self.assertEqual(len(outcome1.storage_read), 0)
        self.assertEqual(len(outcome1.storage_written), 1)
        self.assertBEqual(outcome1.storage_written[utils.bvv(0)], utils.bvv(43))

        self.assertEqual(len(outcome2.storage_read), 0)
        self.assertEqual(len(outcome2.storage_written), 2)
        self.assertBEqual(outcome2.storage_written[utils.bvv(0)], utils.bvv(42))
        self.assertBEqual(outcome2.storage_written[outcome2.env.value], utils.bvv(43))

    def test_sload(self):
        outcome, = self.outcomes([PUSH1, 0, SLOAD, PUSH1, 0, SLOAD])
        self.assertTrue(utils.bvv(0) in outcome.storage_read)
        self.assertEqual(len(outcome.storage_read), 1)

    def test_sstore(self):
        outcome, = self.outcomes(
            [
                PUSH1,
                42,
                PUSH1,
                0,
                SSTORE,
                PUSH1,
                0,
                SLOAD,
                PUSH1,
                43,
                PUSH1,
                0,
                SSTORE,
                PUSH1,
                0,
                SLOAD,
            ]
        )

        self.assertEqual(len(outcome.storage_read), 0)
        self.assertEqual(len(outcome.storage_written), 1)
        self.assertBEqual(outcome.storage_written[utils.bvv(0)], utils.bvv(43))

    # TODO: test_call, test_callcode, test_selfdestruct


if __name__ == "__main__":
    unittest.main()



import logging
import itertools

from claripy.ast import bv
import claripy
import eth_utils


logger = logging.getLogger(__name__)

# TODO: we could keep the resulting constraints in cache or something, so it's
# very fast when we don't use the solver with non-empty extra_constraint.


def Sha3(x):
    return bv.BV("SHA3", [x], length=256)


def _symbolize_hashes(ast, hashes):
    if not isinstance(ast, claripy.ast.base.Base):
        return ast

    # Replace SHA3 with a BVS
    if ast.op == "SHA3":
        hash_input, = ast.args
        hash_input = _symbolize_hashes(hash_input, hashes)
        try:
            return hashes[hash_input]
        except KeyError:
            hash_symbol = claripy.BVS("SHA3", 256)
            hashes[hash_input] = hash_symbol
            logger.debug("Registering new hash: %s(%s)", hash_symbol, hash_input)
            return hash_symbol

    # Recursively apply to children
    args = [_symbolize_hashes(child, hashes) for child in ast.args]
    return ast.swap_args(args)


def _no_sha3_symbol(ast):
    if not isinstance(ast, claripy.ast.base.Base):
        return True
    elif isinstance(ast, claripy.ast.base.BV):
        try:
            return not ast.args[0].startswith("SHA3")
        except AttributeError:
            return True
    else:
        return all(_no_sha3_symbol(child) for child in ast.args)


def _no_sha3_symbols(constraints):
    return all(_no_sha3_symbol(ast) for ast in constraints)


def get_claripy_solver():
    # TODO: What about SolverComposite? Tried, and seems slower.
    return claripy.Solver()


class Solver:
    __slots__ = ["solver", "hashes"]

    def __init__(self, claripy_solver=None, hashes=None):
        self.solver = claripy_solver or get_claripy_solver()
        self.hashes = hashes or {}  # Mapping hash input to the symbol

    def branch(self):
        return Solver(claripy_solver=self.solver.branch(), hashes=self.hashes.copy())

    def add(self, constraints, **kwargs):
        if isinstance(constraints, claripy.ast.base.Base):
            constraints = [constraints]
        logger.debug("Adding constraint: %r", constraints)
        assert _no_sha3_symbols(constraints)
        constraints = [_symbolize_hashes(c, self.hashes) for c in constraints]
        return self.solver.add(constraints, **kwargs)

    def satisfiable(self, extra_constraints=(), **kwargs):
        assert _no_sha3_symbols(extra_constraints)
        try:
            extra_constraints = self._hash_constraints(
                extra_constraints, hashes=self.hashes.copy()
            )
        except claripy.errors.UnsatError:
            return False
        return self.solver.satisfiable(extra_constraints=extra_constraints)

    def eval(self, e, n, extra_constraints=(), **kwargs):
        assert _no_sha3_symbol(e)
        assert _no_sha3_symbols(extra_constraints)
        hashes = self.hashes.copy()
        e = _symbolize_hashes(e, hashes)
        extra_constraints = self._hash_constraints(extra_constraints, hashes=hashes)
        return self.solver.eval(e, n, extra_constraints=extra_constraints)

    def batch_eval(self, e, n, extra_constraints=(), **kwargs):
        raise NotImplementedError()

    def max(self, e, extra_constraints=(), **kwargs):
        assert _no_sha3_symbol(e)
        assert _no_sha3_symbols(extra_constraints)
        hashes = self.hashes.copy()
        e = _symbolize_hashes(e, hashes)
        extra_constraints = self._hash_constraints(extra_constraints, hashes=hashes)
        return self.solver.max(e, extra_constraints=extra_constraints)

    def min(self, e, extra_constraints=(), **kwargs):
        assert _no_sha3_symbol(e)
        assert _no_sha3_symbols(extra_constraints)
        hashes = self.hashes.copy()
        e = _symbolize_hashes(e, hashes)
        extra_constraints = self._hash_constraints(extra_constraints, hashes=hashes)
        return self.solver.min(e, extra_constraints=extra_constraints)

    def solution(self, e, v, extra_constraints=(), **kwargs):
        assert _no_sha3_symbol(e)
        assert _no_sha3_symbols(extra_constraints)
        hashes = self.hashes.copy()
        e = _symbolize_hashes(e, hashes)
        extra_constraints = self._hash_constraints(extra_constraints, hashes=hashes)
        return self.solver.solution(e, v, extra_constraints=extra_constraints)

    def _hash_constraints(self, extra_constraints, hashes, pairs_done=None):
        extra_constraints = [_symbolize_hashes(c, hashes) for c in extra_constraints]

        # Fast-path if no hashes, or if not satisfiable.
        if not hashes or not self.solver.satisfiable(
            extra_constraints=extra_constraints
        ):
            return tuple(extra_constraints)

        if pairs_done is None:
            pairs_done = set()

        constraint_added = False
        for (in1, s1), (in2, s2) in itertools.combinations(hashes.items(), 2):
            if (s1, s2) in pairs_done:
                continue
            # Do s1 needs to be equal to s2 ? Then in1 needs to be equal to in2
            if not self.solver.satisfiable(
                extra_constraints=extra_constraints + [s1 != s2]
            ):
                logger.debug("Adding input constraint: %s == %s", in1, in2)
                if in1.size() == in2.size():
                    extra_constraints.append(in1 == in2)
                else:
                    extra_constraints.append(False)
                pairs_done.add((s1, s2))
                pairs_done.add((s2, s1))
                constraint_added = True
            # Do s1 needs to be != to s2 ? Then in1 needs to be != to in2
            elif not self.solver.satisfiable(
                extra_constraints=extra_constraints + [s1 == s2]
            ):
                logger.debug("Adding input constraint: %s != %s", in1, in2)
                if in1.size() == in2.size():
                    extra_constraints.append(in1 != in2)
                    constraint_added = True
                pairs_done.add((s1, s2))
                pairs_done.add((s2, s1))

        if constraint_added:
            return self._hash_constraints(extra_constraints, hashes, pairs_done)

        assert self.solver.satisfiable(extra_constraints=extra_constraints)

        for in1, s1 in hashes.items():
            # Next line can raise UnsatError. Handled in the caller if needed.
            sol1, = self.solver.eval(in1, 1, extra_constraints=extra_constraints)
            extra_constraints.append(in1 == sol1)
            # lstrip() is needed if the length is 0.
            sol1_bytes = (
                eth_utils.conversions.to_bytes(sol1)
                .lstrip(b"\0")
                .rjust(in1.length // 8, b"\0")
            )
            assert len(sol1_bytes) * 8 == in1.length
            extra_constraints.append(s1 == eth_utils.crypto.keccak(sol1_bytes))
            logger.debug("Added concrete constraint: %s", extra_constraints[-1])

        return tuple(extra_constraints)

    def replace(self, r):
        # First replacement: apply r() everywhere.
        new_constraints = [r(i) for i in self.solver.constraints]
        self.hashes = {r(k): r(v) for k, v in self.hashes.items()}

        # We need to rebuild the solver because we cannot just modify the constraints.
        self.solver = get_claripy_solver()
        self.solver.add(new_constraints)

    def regenerate_hash_symbols(self):
        # We can copy such a solver, and replace symbols in a new environment,
        # and want to combine() it again with the parent solver, or one derived
        # from it. In that case the hashes symbols need to be different! So we
        # can use that function and call replace() on all the symbols to use
        # new hash symbols everywhere.
        new_hashes = {k: claripy.BVS("SHA3", 256) for k, v in self.hashes.items()}

        def r(ast):
            for in_ in self.hashes:
                ast = ast.replace(self.hashes[in_], new_hashes[in_])
            return ast

        return r

    def combine(self, others):
        other_claripy_solvers = [i.solver for i in others]
        combined = Solver(
            claripy_solver=self.solver.combine(other_claripy_solvers),
            hashes=self.hashes,
        )

        for other in others:
            for k, v in self.hashes.items():
                # Make sure the hash symbols are distinct
                if any(v is v2 for v2 in other.hashes.values()):
                    # Call regenerate_hash_symbols() on one of them first?
                    raise ValueError("Cannot combine with equivalent hashes.")

                # TODO: If some hash input are equal, we should merge the hash
                # symols here, it would be more efficient.
            combined.hashes.update(other.hashes)

        return combined

    def downsize(self):
        return self.solver.downsize()

    def simplify(self):
        return self.solver.simplify()

    def __repr__(self):
        return "ClaripySha3(constraints=%s, hashes=%s)" % (
            self.solver.constraints,
            self.hashes,
        )

    def as_dict(self):
        return {"constraints": self.solver.constraints, "hashes": self.hashes}

    @property
    def constraints(self):
        return self.solver.constraints



"""
    pakala: EVM symbolic execution tool and vulnerability scanner.
    Copyright (C) 2018 Korantin Auguste

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import collections
import heapq
import logging
import math
import numbers
import time
import traceback

import claripy

from eth.vm import opcode_values

from pakala import utils
from pakala.state import State
from pakala.claripy_sha3 import Sha3

logger = logging.getLogger(__name__)  # pylint:disable=invalid-name


bvv = utils.bvv  # pylint:disable=invalid-name
BVV_0 = bvv(0)
BVV_1 = bvv(1)

# interesting values aligned to classic parameters.
CALLDATALOAD_INDEX_FUZZ = set(range(0, 32*3, 32)) | set(range(4, 32*5, 32))
CALLDATACOPY_SIZE_FUZZ = set(range(9)) | {0, 32} | {4, 34, 4 + 32 * 16}
RETURNDATACOPY_SIZE_FUZZ = {0, 32}
EXP_EXPONENT_FUZZ = {min, max}


class MultipleSolutionsError(ValueError):
    pass


def bool_to_bv(b):
    return claripy.If(b, BVV_1, BVV_0)


class SymbolicMachine:
    """Class to represent a state of a EVM program, and execute it symbolically.
    """

    def __init__(self, env, fuzz=True):
        self.code = env.code
        logger.debug("Initializing symbolic machine with source code: %s", self.code)
        # For use by heapq only. Contains couples (score, state).
        self.branch_queue = []
        self.states_seen = set()
        self.coverage = [0] * len(self.code)
        # List of all normal/good terminations of the contract
        self.outcomes = []
        # List of all the place where we didn't know how to continue execution
        self.partial_outcomes = []
        # Do we want to enable fuzzing? (see add_for_fuzzing below)
        self.fuzz = fuzz
        # Did fuzzing got used?
        self.fuzzed = False
        # Errors that happened during execution. These are normal.
        self.code_errors = collections.Counter()
        # Errors of the interpreter / symbolic execution engine. Not cool :(
        self.interpreter_errors = collections.Counter()
        self.add_branch(State(env))

    def add_branch(self, state):
        """Add a state corresponding to a branch to be executed."""
        if not state.solver.satisfiable():
            logger.debug("Avoided adding unsatisfiable state.")
            return

        if hash(state) in self.states_seen:
            logger.debug("Avoided adding visited state.")
            self.code_errors["Avoided adding visited state"] += 1
            return

        # We have to downsize the used solver to free memory in z3, otherwise
        # they will collectively end up eating all the memory.
        state.solver.downsize()
        state.solver.simplify()

        logger.debug(
            "Adding branch to %i with depth %i (visited %i times)",
            state.pc,
            state.depth,
            self.coverage[state.pc] if state.pc < len(self.coverage) else 0,
        )

        heapq.heappush(self.branch_queue, (state.depth, state))
        self.states_seen.add(hash(state))

    def add_for_fuzzing(self, state, variable, tries):
        """
        Will try to fuzz the variable, setting it to different values.
        The tries parameter must be an array of :
         - min: if you want to add the minimum value possible
         - max: if you want to add the maximum value possible
         - a number: to try that number
         - None: a random number that works, you can repeat it.
        """
        # TODO: If the fuzzer is used, then I will generate tons of branches
        # that are equivalent... There should be a way to deduplicate them,
        # but it's not trivial.

        if not self.fuzz:
            raise utils.InterpreterError(state, "Fuzzer is disabled")
        if not self.fuzzed:
            self.fuzzed = True
            logger.warning("Fuzzer got used (forced concretization). "
                           "We will lose accuracy and risk state explosion.")

        to_try = set()
        nb_random = 0
        for t in tries:  # pylint:disable=invalid-name
            if isinstance(t, numbers.Number) and state.solver.solution(variable, t):
                to_try.add(t)
            elif t is min:
                to_try.add(state.solver.min(variable))
            elif t is max:
                to_try.add(state.solver.max(variable))
            elif t is None:
                nb_random += 1
        if nb_random:
            to_try |= set(state.solver.eval(variable, nb_random))

        logger.debug("Fuzzing will try %s in %s.", variable, to_try)
        state.depth += 1 if len(to_try) == 1 else 10  # Lower the priority of what we got by fuzzing.
        for value in to_try:
            new_state = state.copy()
            new_state.solver.add(variable == value)
            self.add_branch(new_state)

    def exec_branch(self, state):  # pylint:disable=invalid-name
        """Execute forward from a state, queuing new states if needed."""
        logger.debug("Constraints: %s", state.solver.constraints)

        def solution(variable):
            """Returns the solution. There must be one or we fail."""
            solutions = state.solver.eval(variable, 2)
            if len(solutions) > 1:
                raise MultipleSolutionsError(
                    "Multiple solutions for %s (%#x)" % (variable, self.code[state.pc])
                )
            solution = solutions[0]
            return solution if isinstance(solution, numbers.Number) else solution.value

        self.code.pc = state.pc

        while True:
            if state.pc >= len(self.code):
                return True

            op = next(self.code)
            self.coverage[state.pc] += 1

            logger.debug("NEW STEP")
            logger.debug("Memory: %s", state.memory)
            logger.debug("Stack: %s", state.stack)
            logger.debug("PC: %i, %#x", state.pc, op)

            assert self.code.pc == state.pc + 1
            assert isinstance(op, numbers.Number)
            assert all(
                isinstance(i, claripy.ast.base.BV) for i in state.stack
            ), "The stack musty only contains claripy BV's"

            # Trivial operations first
            if not self.code.is_valid_opcode(state.pc):
                raise utils.CodeError("Trying to execute PUSH data")
            elif op == 254:  # INVALID opcode
                raise utils.CodeError("designed INVALID opcode")
            elif op == opcode_values.JUMPDEST:
                pass
            elif op == opcode_values.ADD:
                s0, s1 = (
                    state.stack_pop(),
                    state.stack_pop(),
                )  # pylint:disable=invalid-name
                state.stack_push(s0 + s1)
            elif op == opcode_values.SUB:
                s0, s1 = (
                    state.stack_pop(),
                    state.stack_pop(),
                )  # pylint:disable=invalid-name
                state.stack_push(s0 - s1)
            elif op == opcode_values.MUL:
                s0, s1 = (
                    state.stack_pop(),
                    state.stack_pop(),
                )  # pylint:disable=invalid-name
                state.stack_push(s0 * s1)
            elif op == opcode_values.DIV:
                # We need to use claripy.LShR instead of a division if possible,
                # because the solver is bad dealing with divisions, better
                # with shifts. And we need shifts to handle the solidity ABI
                # for function selection.
                s0, s1 = (
                    state.stack_pop(),
                    state.stack_pop(),
                )  # pylint:disable=invalid-name
                try:
                    s1 = solution(s1)  # pylint:disable=invalid-name
                except MultipleSolutionsError:
                    state.stack_push(claripy.If(s1 == 0, BVV_0, s0 / s1))
                else:
                    if s1 == 0:
                        state.stack_push(BVV_0)
                    elif s1 == 1:
                        state.stack_push(s0)
                    elif s1 & (s1 - 1) == 0:
                        exp = int(math.log(s1, 2))
                        state.stack_push(s0.LShR(exp))
                    else:
                        state.stack_push(s0 / s1)
            elif op == opcode_values.SDIV:
                s0, s1 = (
                    state.stack_pop(),
                    state.stack_pop(),
                )  # pylint:disable=invalid-name
                try:
                    s1 = solution(s1)
                except MultipleSolutionsError:
                    state.stack_push(claripy.If(s1 == 0, BVV_0, s0.SDiv(s1)))
                else:
                    state.stack_push(BVV_0 if s1 == 0 else s0.SDiv(s1))
            elif op == opcode_values.MOD:
                s0, s1 = (
                    state.stack_pop(),
                    state.stack_pop(),
                )  # pylint:disable=invalid-name
                try:
                    s1 = solution(s1)
                except MultipleSolutionsError:
                    state.stack_push(claripy.If(s1 == 0, BVV_0, s0 % s1))
                else:
                    state.stack_push(BVV_0 if s1 == 0 else s0 % s1)
            elif op == opcode_values.SMOD:
                s0, s1 = (
                    state.stack_pop(),
                    state.stack_pop(),
                )  # pylint:disable=invalid-name
                try:
                    s1 = solution(s1)
                except MultipleSolutionsError:
                    state.stack_push(claripy.If(s1 == 0, BVV_0, s0.SMod(s1)))
                else:
                    state.stack_push(BVV_0 if s1 == 0 else s0.SMod(s1))
            elif op == opcode_values.ADDMOD:
                s0, s1, s2 = state.stack_pop(), state.stack_pop(), state.stack_pop()
                try:
                    s2 = solution(s2)
                except MultipleSolutionsError:
                    state.stack_push(claripy.If(s2 == 0, BVV_0, (s0 + s1) % s2))
                else:
                    state.stack_push(BVV_0 if s2 == 0 else (s0 + s1) % s2)
            elif op == opcode_values.MULMOD:
                s0, s1, s2 = state.stack_pop(), state.stack_pop(), state.stack_pop()
                try:
                    s2 = solution(s2)
                except MultipleSolutionsError:
                    state.stack_push(claripy.If(s2 == 0, BVV_0, (s0 * s1) % s2))
                else:
                    state.stack_push(BVV_0 if s2 == 0 else (s0 * s1) % s2)
            elif op == opcode_values.SHL:
                shift, value = state.stack_pop(), state.stack_pop()
                state.stack_push(value << shift)
            elif op == opcode_values.SHR:
                shift, value = state.stack_pop(), state.stack_pop()
                state.stack_push(value.LShR(shift))
            elif op == opcode_values.SAR:
                shift, value = state.stack_pop(), state.stack_pop()
                state.stack_push(claripy.RotateRight(value, shift))
            elif op == opcode_values.EXP:
                base, exponent = state.stack_pop(), state.stack_pop()
                base_sol = solution(base)
                if base_sol == 2:
                    state.stack_push(1 << exponent)
                else:
                    try:
                        exponent_sol = solution(exponent)
                    except MultipleSolutionsError:
                        state.stack_push(exponent)  # restore stack
                        state.stack_push(base)
                        self.add_for_fuzzing(state, exponent, EXP_EXPONENT_FUZZ)
                        return False
                    else:
                        state.stack_push(claripy.BVV(base_sol ** exponent_sol, 256))
            elif op == opcode_values.LT:
                s0, s1 = (
                    state.stack_pop(),
                    state.stack_pop(),
                )  # pylint:disable=invalid-name
                state.stack_push(bool_to_bv(claripy.ULT(s0, s1)))
            elif op == opcode_values.GT:
                s0, s1 = (
                    state.stack_pop(),
                    state.stack_pop(),
                )  # pylint:disable=invalid-name
                state.stack_push(bool_to_bv(claripy.UGT(s0, s1)))
            elif op == opcode_values.SLT:
                s0, s1 = (
                    state.stack_pop(),
                    state.stack_pop(),
                )  # pylint:disable=invalid-name
                state.stack_push(bool_to_bv(claripy.SLT(s0, s1)))
            elif op == opcode_values.SGT:
                s0, s1 = (
                    state.stack_pop(),
                    state.stack_pop(),
                )  # pylint:disable=invalid-name
                state.stack_push(bool_to_bv(claripy.SGT(s0, s1)))
            elif op == opcode_values.SIGNEXTEND:
                # TODO: Use Claripy's SignExt that should do exactly that.
                s0, s1 = (
                    state.stack_pop(),
                    state.stack_pop(),
                )  # pylint:disable=invalid-name
                # s0 is the number of bits. s1 the number we want to extend.
                s0 = solution(s0)
                if s0 <= 31:
                    sign_bit = 1 << (s0 * 8 + 7)
                    state.stack_push(
                        claripy.If(
                            s1 & sign_bit == 0,
                            s1 & (sign_bit - 1),
                            s1 | ((1 << 256) - sign_bit),
                        )
                    )
                else:
                    state.stack_push(s1)
            elif op == opcode_values.EQ:
                s0, s1 = state.stack_pop(), state.stack_pop()
                state.stack_push(bool_to_bv(s0 == s1))
            elif op == opcode_values.ISZERO:
                state.stack_push(bool_to_bv(state.stack_pop() == BVV_0))
            elif op == opcode_values.AND:
                s0, s1 = state.stack_pop(), state.stack_pop()
                state.stack_push(s0 & s1)
            elif op == opcode_values.OR:
                s0, s1 = state.stack_pop(), state.stack_pop()
                state.stack_push(s0 | s1)
            elif op == opcode_values.XOR:
                s0, s1 = state.stack_pop(), state.stack_pop()
                state.stack_push(s0 ^ s1)
            elif op == opcode_values.NOT:
                state.stack_push(~state.stack_pop())
            elif op == opcode_values.BYTE:
                s0, s1 = (
                    state.stack_pop(),
                    state.stack_pop(),
                )  # pylint:disable=invalid-name
                state.stack_push(s1.LShR(claripy.If(s0 > 31, 32, 31 - s0) * 8) & 0xFF)

            elif op == opcode_values.PC:
                state.stack_push(bvv(state.pc))
            elif op == opcode_values.GAS:
                state.stack_push(state.env.gas)
            elif op == opcode_values.ADDRESS:
                state.stack_push(state.env.address)
            elif op == opcode_values.BALANCE:
                addr = solution(state.stack_pop())
                if addr != solution(state.env.address):
                    raise utils.InterpreterError(
                        state, "Can only query balance of the current contract for now"
                    )
                state.stack_push(state.env.balance)
            elif op == opcode_values.ORIGIN:
                state.stack_push(state.env.origin)
            elif op == opcode_values.CALLER:
                state.stack_push(state.env.caller)
            elif op == opcode_values.CALLVALUE:
                state.stack_push(state.env.value)
            elif op == opcode_values.BLOCKHASH:
                block_num = state.stack_pop()
                if block_num not in state.env.block_hashes:
                    state.env.block_hashes[block_num] = claripy.BVS(
                        "blockhash[%s]" % block_num, 256
                    )
                state.stack_push(state.env.block_hashes[block_num])
            elif op == opcode_values.TIMESTAMP:
                state.stack_push(state.env.block_timestamp)
            elif op == opcode_values.NUMBER:
                state.stack_push(state.env.block_number)
            elif op == opcode_values.COINBASE:
                state.stack_push(state.env.coinbase)
            elif op == opcode_values.DIFFICULTY:
                state.stack_push(state.env.difficulty)
            elif op == opcode_values.POP:
                state.stack_pop()
            elif op == opcode_values.JUMP:
                addr = solution(state.stack_pop())
                if addr >= len(self.code) or self.code[addr] != opcode_values.JUMPDEST:
                    raise utils.CodeError("Invalid jump (%i)" % addr)
                state.pc = addr
                self.add_branch(state)
                return False
            elif op == opcode_values.JUMPI:
                addr, condition = solution(state.stack_pop()), state.stack_pop()
                state_false = state.copy()
                state.solver.add(condition != BVV_0)
                state_false.solver.add(condition == BVV_0)
                state_false.pc += 1
                self.add_branch(state_false)
                state.pc = addr
                if (
                    state.pc >= len(self.code)
                    or self.code[state.pc] != opcode_values.JUMPDEST
                ):
                    raise utils.CodeError("Invalid jump (%i)" % (state.pc - 1))
                self.add_branch(state)
                return False
            elif opcode_values.PUSH1 <= op <= opcode_values.PUSH32:
                pushnum = op - opcode_values.PUSH1 + 1
                raw_value = self.code.read(pushnum)
                state.pc += pushnum
                state.stack_push(bvv(int.from_bytes(raw_value, byteorder="big")))
            elif opcode_values.DUP1 <= op <= opcode_values.DUP16:
                depth = op - opcode_values.DUP1 + 1
                state.stack_push(state.stack[-depth])
            elif opcode_values.SWAP1 <= op <= opcode_values.SWAP16:
                depth = op - opcode_values.SWAP1 + 1
                temp = state.stack[-depth - 1]
                state.stack[-depth - 1] = state.stack[-1]
                state.stack[-1] = temp
            elif opcode_values.LOG0 <= op <= opcode_values.LOG4:
                depth = op - opcode_values.LOG0
                mstart, msz = (state.stack_pop(), state.stack_pop())
                topics = [state.stack_pop() for x in range(depth)]
            elif op == opcode_values.SHA3:
                start, length = solution(state.stack_pop()), solution(state.stack_pop())
                memory = state.memory.read(start, length)
                state.stack_push(Sha3(memory))
            elif op == opcode_values.STOP:
                return True
            elif op == opcode_values.RETURN:
                return True

            elif op == opcode_values.CALLDATALOAD:
                index = state.stack_pop()
                try:
                    index_sol = solution(index)
                except MultipleSolutionsError:
                    state.stack_push(index)  # restore the stack
                    self.add_for_fuzzing(state, index, CALLDATALOAD_INDEX_FUZZ)
                    return False
                state.stack_push(state.env.calldata.read(index_sol, 32))
            elif op == opcode_values.CALLDATASIZE:
                state.stack_push(state.env.calldata_size)
            elif op == opcode_values.CALLDATACOPY:
                old_state = state.copy()
                mstart, dstart, size = (
                    state.stack_pop(),
                    state.stack_pop(),
                    state.stack_pop(),
                )
                mstart, dstart = solution(mstart), solution(dstart)
                try:
                    size = solution(size)
                except MultipleSolutionsError:
                    self.add_for_fuzzing(old_state, size, CALLDATACOPY_SIZE_FUZZ)
                    return False
                state.memory.copy_from(state.env.calldata, mstart, dstart, size)
            elif op == opcode_values.CODESIZE:
                state.stack_push(bvv(len(self.code)))
            elif op == opcode_values.EXTCODESIZE:
                addr = state.stack_pop()
                if (addr == state.env.address).is_true():
                    state.stack_push(bvv(len(self.code)))
                else:
                    # TODO: Improve that... It's clearly not constraining enough.
                    state.stack_push(claripy.BVS("EXTCODESIZE[%s]" % addr, 256))

            elif op == opcode_values.EXTCODECOPY:
                old_state = state.copy()
                addr = state.stack_pop()
                mem_start = solution(state.stack_pop())
                code_start = solution(state.stack_pop())

                size = state.stack_pop()
                try:
                    size = solution(size)
                except MultipleSolutionsError:
                    # TODO: Fuzz.
                    # self.add_for_fuzzing(old_state, size, [])
                    # return False
                    raise
                state.memory.write(
                    mem_start,
                    size,
                    claripy.BVS("EXTCODE[%s from %s]" % (addr, code_start), size * 8),
                )

            elif op == opcode_values.CODECOPY:
                mem_start, code_start, size = [
                    solution(state.stack_pop()) for _ in range(3)
                ]
                for i in range(size):
                    if code_start + i < len(state.env.code):
                        state.memory.write(
                            mem_start + i,
                            1,
                            claripy.BVV(state.env.code[code_start + i], 8),
                        )
                    else:
                        state.memory.write(mem_start + i, 1, claripy.BVV(0, 8))

            elif op == opcode_values.MLOAD:
                index = solution(state.stack_pop())
                state.stack_push(state.memory.read(index, 32))
            elif op == opcode_values.MSTORE:
                index, value = solution(state.stack_pop()), state.stack_pop()
                state.memory.write(index, 32, value)
            elif op == opcode_values.MSTORE8:
                index, value = solution(state.stack_pop()), state.stack_pop()
                state.memory.write(index, 1, value[7:0])
            elif op == opcode_values.MSIZE:
                state.stack_push(bvv(state.memory.size()))

            elif op == opcode_values.SLOAD:
                state.pc += 1
                key = state.stack_pop()
                for w_key, w_value in state.storage_written.items():
                    read_written = [w_key == key]
                    if state.solver.satisfiable(extra_constraints=read_written):
                        new_state = state.copy()
                        new_state.solver.add(read_written)
                        new_state.stack_push(w_value)
                        self.add_branch(new_state)
                        state.solver.add(w_key != key)
                if state.solver.satisfiable():
                    assert key not in state.storage_written
                    if key not in state.storage_read:
                        state.storage_read[key] = claripy.BVS("storage[%s]" % key, 256)
                    state.stack_push(state.storage_read[key])
                    self.add_branch(state)
                return

            elif op == opcode_values.SSTORE:
                state.pc += 1
                key = state.stack_pop()
                value = state.stack_pop()
                for w_key, w_value in state.storage_written.items():
                    read_written = [w_key == key]
                    if state.solver.satisfiable(extra_constraints=read_written):
                        new_state = state.copy()
                        new_state.solver.add(read_written)
                        new_state.storage_written[w_key] = value
                        self.add_branch(new_state)
                        state.solver.add(w_key != key)
                if state.solver.satisfiable():
                    assert key not in state.storage_written
                    state.storage_written[key] = value
                    self.add_branch(state)
                return

            elif op == opcode_values.CALL:
                state.pc += 1

                # pylint:disable=unused-variable
                gas, to_, value, meminstart, meminsz, memoutstart, memoutsz = (
                    state.stack_pop() for _ in range(7)
                )

                # First possibility: the call fails
                # (always possible with a call stack big enough)
                state_fail = state.copy()
                state_fail.stack_push(BVV_0)
                self.add_branch(state_fail)

                # Second possibility: success.
                state.calls.append(
                    (memoutsz, memoutstart, meminsz, meminstart, value, to_, gas)
                )

                memoutsz = solution(memoutsz)
                if memoutsz != 0:
                    # If we expect some output, let's constraint the call to
                    # be to a contract that we do control. Otherwise it could
                    # return anything...
                    state.solver.add(to_[159:0] == utils.DEFAULT_CALLER[159:0])

                    memoutstart = solution(memoutstart)
                    state.memory.write(
                        memoutstart,
                        memoutsz,
                        claripy.BVS("CALL_RETURN[%s]" % to_, memoutsz * 8),
                    )

                state.stack_push(BVV_1)
                self.add_branch(state)
                return False

            elif op == opcode_values.DELEGATECALL:
                state.pc += 1

                # pylint:disable=unused-variable
                gas, to_, meminstart, meminsz, memoutstart, memoutsz = (
                    state.stack_pop() for _ in range(6)
                )

                # First possibility: the call fails
                # (always possible with a call stack big enough)
                state_fail = state.copy()
                state_fail.stack_push(BVV_0)
                self.add_branch(state_fail)

                # If the call is to a specific contract we don't control,
                # don't assume it could return anything, or even be successful.
                # So we say we need to be able to call an arbitrary contract.
                state.solver.add(to_[159:0] == utils.DEFAULT_CALLER[159:0])

                # Second possibility: success.
                state.calls.append(
                    (memoutsz, memoutstart, meminsz, meminstart, to_, gas)
                )

                memoutsz = solution(memoutsz)
                if memoutsz != 0:
                    memoutstart = solution(memoutstart)
                    state.memory.write(
                        memoutstart,
                        memoutsz,
                        claripy.BVS("DELEGATECALL_RETURN[%s]" % to_, memoutsz * 8),
                    )

                state.stack_push(BVV_1)
                self.add_branch(state)
                return False

            elif op == opcode_values.RETURNDATASIZE:
                state.stack_push(claripy.BVS("RETURNDATASIZE", 256))

            elif op == opcode_values.RETURNDATACOPY:
                old_state = state.copy()
                mem_start_position = solution(state.stack_pop())
                returndata_start_position = solution(state.stack_pop())

                size = state.stack_pop()
                try:
                    size = solution(size)
                except MultipleSolutionsError:
                    self.add_for_fuzzing(old_state, size, RETURNDATACOPY_SIZE_FUZZ)
                    return False

                state.memory.write(
                    mem_start_position, size, claripy.BVS("RETURNDATACOPY", size * 8)
                )

            elif op == opcode_values.SELFDESTRUCT:
                state.selfdestruct_to = state.stack[-1]
                return True

            elif op == opcode_values.REVERT:
                return False

            else:
                raise utils.InterpreterError(state, "Unknown opcode %#x" % op)

            state.pc += 1

    def execute(self, timeout_sec):
        """Run the code, searching for all the interesting outcomes.

        Returns the process time it took to execute.
        """

        if self.outcomes:
            raise RuntimeError("Already executed.")

        time_start = time.process_time()

        time_last_coverage_increase = time_start
        last_coverage = 0

        while self.branch_queue:

            coverage = sum(bool(c) for c in self.coverage)
            if coverage > last_coverage:
                if time.process_time() - time_last_coverage_increase > 1:
                    logger.log(
                        utils.INFO_INTERACTIVE,
                        "Coverage now %i%%. Queue size %i. Got %i outcomes.",
                        int(self.get_coverage() * 100),
                        len(self.branch_queue),
                        len(self.outcomes),
                    )
                time_last_coverage_increase = time.process_time()
                last_coverage = coverage

            if (
                not timeout_sec
                and time.process_time() - time_last_coverage_increase
                > max(120, time_last_coverage_increase - time_start)
            ) or (timeout_sec and time.process_time() - time_start > timeout_sec):
                logger.debug("Timeout.")
                self.interpreter_errors["execute timeout"] += 1
                break

            depth, state = heapq.heappop(self.branch_queue)
            state.depth += 1

            logger.debug("Executing branch at %i with depth %i.", state.pc, depth)
            try:
                success = self.exec_branch(state)
            except KeyboardInterrupt:
                self.interpreter_errors['KeyboardInterrupt'] += 1
                break
            except (utils.CodeError, claripy.errors.UnsatError) as error:
                logger.debug("Code error: %s", error)
                self.code_errors[repr(error)] += 1
            except (
                utils.InterpreterError,
                claripy.errors.ClaripyError,
                MultipleSolutionsError,
                ZeroDivisionError,
            ) as error:
                logger.debug("Interpreter error: %s", error)
                logger.debug(traceback.format_exc())
                self.interpreter_errors[repr(error)] += 1
                if isinstance(error, utils.InterpreterError):
                    self.add_partial_outcome(error.state)
            else:
                if success:
                    self.add_outcome(state)
                logger.debug("Branch done.")

        # In case of timeouts, we still have unfinished branches in the queue!
        # Add them as partial outcomes.
        while self.branch_queue:
            depth, state = heapq.heappop(self.branch_queue)
            self.add_partial_outcome(state)

        logger.info(
            "Analysis finished with %i outcomes (%i interesting, %i unfinished), "
            "coverage is %i%%",
            len(self.outcomes),
            sum(int(o.is_interesting()) for o in self.outcomes),
            len(self.partial_outcomes),
            int(self.get_coverage() * 100),
        )

        if self.code_errors:
            logger.info("Code errors encountered: %s", self.code_errors.most_common())
        if self.interpreter_errors:
            logger.info(
                "Interpreter errors encountered: %s",
                self.interpreter_errors.most_common(),
            )

        logger.debug("List of outcomes:")
        for outcome in self.outcomes:
            logger.debug(outcome.debug_string())

    def add_outcome(self, state):
        """Add an outcome to the list."""
        state.clean()
        logger.debug("Adding outcome: %s", state.debug_string())
        self.outcomes.append(state)

    def add_partial_outcome(self, state):
        """Add an outcome to the list of partial outcomes."""
        state.clean()
        logger.debug("Adding partial outcome: %s", state.debug_string())
        self.partial_outcomes.append(state)

    def get_coverage(self):
        """Return the ratio of instructions that were executed by the total
        number of instructions."""
        logger.debug("Coverage analysis:")
        total_lines = 0
        covered_lines = 0
        self.code.pc = 0
        for pc, instruction in enumerate(self.code):  # pylint:disable=invalid-name
            if pc == len(self.code):
                break
            if not self.code.is_valid_opcode(pc):
                continue
            if instruction == opcode_values.JUMPDEST:
                logger.debug(
                    "  {:04x}: {}".format(
                        pc, "covered" if self.coverage[pc] else "not covered"
                    )
                )
            total_lines += 1
            covered_lines += bool(self.coverage[pc])
        return covered_lines / float(total_lines or 1)



"""
    pakala: EVM symbolic execution tool and vulnerability scanner.
    Copyright (C) 2018 Korantin Auguste

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import argparse
import codecs
import logging
import sys
import re

from pakala import sm
from pakala import recursive_analyzer
from pakala import env
from pakala import utils
from pakala import summary

from web3.auto import w3
from web3 import Web3
import web3
import coloredlogs


def err_exit(message):
    print(message, file=sys.stderr)
    exit(-1)


def ethWeiAmount(arg):
    m = re.match(r"^([0-9.]+) ?(\w+)$", arg)
    if m is None:
        raise argparse.ArgumentError(
            "The argument must be in the form '1 ether' for example."
        )
    return Web3.toWei(float(m.group(1)), m.group(2))


def addressOrStdin(s):
    if s == "-":
        return s
    if not re.match(r"^0x([0-9a-fA-F]){40}$", s):
        raise argparse.ArgumentError("Invalid address.")
    return Web3.toChecksumAddress(s)


parser = argparse.ArgumentParser(
    description="Find exploitable Ethereum smart contracts."
)

parser.add_argument(
    "contract_addr",
    type=addressOrStdin,
    help="Address of the contract to analyze. "
    "Use '-' for reading runtime bytecode from stdin instead.",
)
parser.add_argument(
    "-v",
    default=str(utils.INFO_INTERACTIVE),
    help="log level (INFO, DEBUG...)",
    metavar="LOG_LEVEL",
)
parser.add_argument(
    "-s", "--summarize", action="store_true", help="enable summarizer (EXPERIMENTAL)"
)

limits = parser.add_argument_group("time/depth limits")
limits.add_argument(
    "--exec-timeout",
    help=(
        "Timeout in seconds for the symbolic execution stage. Use 0 for a "
        "system that will stop when the last coverage increase was too long ago."
    ),
    type=int,
    default=0,
    metavar="SECONDS",
)
limits.add_argument(
    "--analysis-timeout",
    help=(
        "Timeout in seconds for the analysis stage (that will stack the executions "
        "and find bugs). Use 0 to disable timeout and use only depth limit."
    ),
    type=int,
    default=0,
    metavar="SECONDS",
)
limits.add_argument(
    "--max-transaction-depth",
    help=(
        "Maximum number of outcomes that can be fused "
        "together during the analysis step."
    ),
    type=int,
    default=4,
)


environment = parser.add_argument_group("environment")
environment.add_argument(
    "-b",
    "--force-balance",
    type=ethWeiAmount,
    help="Don't use the current contract balance, instead force it to a value.",
    metavar="BALANCE",
)
environment.add_argument(
    "-B",
    "--block",
    default="latest",
    type=lambda block_number: hex(int(block_number))
    if block_number.isnumeric()
    else block_number,
    help="Use the code/balance/storage at that block instead of latest.",
)

symbolic = parser.add_argument_group("symbolic execution tweaks")
symbolic.add_argument(
    "-z",
    "--disable-fuzzing",
    action="store_true",
    help="Disable forced concretization of symbols where we need a concrete value.",
)

analyzer = parser.add_argument_group("analyzer tweaks")
analyzer.add_argument(
    "-m",
    "--min-to-receive",
    type=ethWeiAmount,
    default="1 milliether",
    help="Minimum amount to receive from the contract to consider it a bug.",
    metavar="BALANCE",
)
analyzer.add_argument(
    "-M",
    "--max-to-send",
    type=ethWeiAmount,
    default="10 ether",
    help=(
        "Maximum amount allowed to send to the contract "
        "(even if we would receive more)."
    ),
    metavar="BALANCE",
)


def main():
    args = parser.parse_args()

    if args.v.isnumeric():
        coloredlogs.install(level=int(args.v))
    elif hasattr(logging, args.v.upper()):
        coloredlogs.install(level=getattr(logging, args.v.upper()))
    else:
        err_exit("Logging should be DEBUG/INFO/WARNING/ERROR.")

    try:
        logging.debug("Node working. Block %i ", w3.eth.blockNumber)
    except web3.exceptions.UnhandledRequest:
        err_exit(
            "Seems like Web3.py can't connect to your Ethereum node.\n"
            "If you don't have one and you want to use Infura, you can set "
            "WEB3_PROVIDER_URI as follows:\n"
            "$ export WEB3_PROVIDER_URI='https://mainnet.infura.io'"
        )

    if args.contract_addr == "-":
        # Let's read the runtime bytecode from stdin
        code = sys.stdin.read().strip("\n")
        if not code.isalnum():
            err_exit("Runtime bytecode read from stdin needs to be hexadecimal.")
        code = codecs.decode(code, "hex")
        # Dummy address, dummy balance
        args.contract_addr = "0xDEADBEEF00000000000000000000000000000000"
        if not args.force_balance:
            args.force_balance = Web3.toWei(1.337, "ether")
    else:
        code = w3.eth.getCode(args.contract_addr, block_identifier=args.block)

    balance = args.force_balance or w3.eth.getBalance(
        args.contract_addr, block_identifier=args.block
    )

    print(
        "Analyzing contract at %s with balance %f ether."
        % (args.contract_addr, Web3.fromWei(balance, "ether"))
    )

    if balance < args.min_to_receive:
        err_exit(
            "Balance is smaller than --min-to-receive: "
            "the analyzer will never find anything."
        )

    if args.summarize:
        logging.info(
            "Summarizer enabled, we won't constrain the caller/origin "
            "so more of the contract can get explored. "
            "It may be slower."
        )
        e = env.Env(
            code,
            address=utils.bvv(int(args.contract_addr, 16)),
            balance=utils.bvv(balance),
        )
    else:
        e = env.Env(
            code,
            address=utils.bvv(int(args.contract_addr, 16)),
            caller=utils.DEFAULT_CALLER,
            origin=utils.DEFAULT_CALLER,
            balance=utils.bvv(balance),
        )

    print("Starting symbolic execution step...")

    s = sm.SymbolicMachine(e, fuzz=not args.disable_fuzzing)
    s.execute(timeout_sec=args.exec_timeout)

    print(
        "Symbolic execution finished with coverage %i%%." % int(s.get_coverage() * 100)
    )
    print(
        "Outcomes: %i interesting. %i total and %i unfinished paths."
        % (
            sum(int(o.is_interesting()) for o in s.outcomes),
            len(s.outcomes),
            len(s.partial_outcomes),
        )
    )

    if args.summarize:
        print()
        print("Methods from the summarizer:")
        summary.HumanSummarizer(s).print_methods()

    print()
    print("Starting analysis step...")

    ra = recursive_analyzer.RecursiveAnalyzer(
        max_wei_to_send=args.max_to_send,
        min_wei_to_receive=args.min_to_receive,
        block=args.block,
    )
    bug = ra.check_states(
        s.outcomes, timeout=args.analysis_timeout, max_depth=args.max_transaction_depth
    )

    if bug:
        solver = bug[2]
        if logging.getLogger().isEnabledFor(logging.DEBUG):
            print("Composite state:")
            print(bug[0].debug_string())
            print()
            print()
        print("Path:")
        for i, state in enumerate(bug[1]):
            print()
            print("Transaction %i, symbolic state:" % (i + 1))
            print(state.debug_string())
            print()
            print("Transaction %i, example solution:" % (i + 1))
            print(state.env.solution_string(solver))
            print()
        print()
        print("======> Bug found! Need %i transactions. <======" % len(bug[1]))
    else:
        print("Nothing to report.")


if __name__ == "__main__":
    main()
